// This program is used to generate artificial dataset.

#include <fstream>
#include <cmath>
#include <iostream>

#include "global.h"
#include "header.h"
#include "utils-file.h"
#include "utils-string.h"

#define ranf() ((double)random()/(1.0+(double)RAND_MAX))
#define MODMULT(a,b,c,m,s) q=s/a; s=b*(s-a*q)-c*q; if (s<0) s+=m;

using namespace std;

int cov;
int ins_len;
int low_len;
int high_len;
int seed_len;
int seed_num;
int std_dev;

long s1 = -1;
long s2 = -1;

// There are for the random number generator
//////////////////////
double norm = 2.3283163396834613e-10;
double m1   = 4294949027.0;
double m2   = 4294934327.0;
double a12  =    1154721.0;
double a14  =    1739991.0;
double a15n =    1108499.0;
double a21  =    1776413.0;
double a23  =     865203.0;
double a25n =    1641052.0;
double s10  = -1;
double s11  = -1;
double s12  = -1;
double s13  = -1;
double s14  = -1;
double s20  = -1;
double s21  = -1;
double s22  = -1;
double s23  = -1;
double s24  = -1;
//////////////////////

string distro;
string machine;
string org;
string orgrc;

fstream* forg;
fstream* fpair;
fstream* fseed;

void GenSeed(int seed_len, int num_seed);
void GenABDat();

int main(int argc, char** argv) {
  if (argc != 11) {
    cout << "usage: org-seq [ab/helicos] read-len-low read-len-high read-coverage seed-len "
      "seed-num ins-len std-dev [normal/random]\n";
    exit(1);
  }
  forg = UtilsFile::OpenFile(argv[1], "r");
  (*forg) >> org;
  orgrc = UtilsString::ReverseComplement(org);
  machine = argv[2];
  low_len = atoi(argv[3]);
  high_len = atoi(argv[4]);
  cov = atoi(argv[5]);
  seed_len = atoi(argv[6]);
  seed_num = atoi(argv[7]);
  ins_len = atoi(argv[8]);
  std_dev = atoi(argv[9]);
  distro = argv[10];
  fpair = UtilsFile::OpenFile("out/pair" + UtilsString::Int2Str(cov) + "_" +
			      UtilsString::Int2Str(ins_len) + "_" +
			      UtilsString::Int2Str(std_dev) + ".fasta", "w");
  fseed = UtilsFile::OpenFile("out/seed.fasta", "w");

  GenSeed(seed_len, seed_num);

  if (machine == "ab") {
    GenABDat();
  }

  UtilsFile::CloseFile(forg);
  UtilsFile::CloseFile(fpair);
  UtilsFile::CloseFile(fseed);
}

string AddErr2AB(const string& str) {
  //return str;
  char prev = str[0];
  //  cout << "before: " << str << endl;
  string mar = UtilsString::Fasta2Solid(str);
  //cout << "after: " << mar << endl;
  //  int x = rand() % 100;
  //int err = 0;

  //37.88(0)   27.36(1)  20.12(2) 14.63(3) 0
  // if (x <= 27) err = 1;
  //else if (x <= 48) err = 2;
  //else if (x <= 63) err = 3;
  
  // int pos[3];

  // Pick the random positions.
  // for (int i = 0; i < err; ++i) {
  //  pos[i] = rand() % mar.size();
  // }

  // Contaminate the positions.
  for (int i = 0; i < mar.size(); ++i) {
    if (rand() % 1000 != 6) continue;
    char c = UtilsString::Int2Base(rand() % 4);
    while (c == mar[i]) {
      c = UtilsString::Int2Base(rand() % 4);
    }
    mar[i] = c;
  }
  //  for (int i = 0; i < mar.size(); ++i) {
  //  mar[i] = UtilsString::Base2Int(mar[i]) + '0';
  // }
  string t = "";
  t.push_back(prev);
  mar.insert(0, t);
  //mar = UtilsString::Solid2Fasta(mar);
  return mar;
}

double SimpleUniform() {
  return (double)rand() / RAND_MAX;
}

double ecuyer0() {
  long q,z;
  if (s1 < 0) {
    s1 = long(2147483563L * SimpleUniform());
    s2 = long(2147483399L * SimpleUniform());
  }
  MODMULT(53668,40014,12211,2147483563L,s1);
  MODMULT(52774,40692,3791,2147483399L,s2);
  z = s1 - s2;
  if (z < 1)
    z += 2147483562;
  return z * 4.65661305956e-10;
}

double ecuyer2() {
  long   k;
  double d,p1, p2;
  if (s10 < 0) {
    s10 = floor(m1 * SimpleUniform());
    s11 = floor(m1 * SimpleUniform());
    s12 = floor(m1 * SimpleUniform());
    s13 = floor(m1 * SimpleUniform());
    s14 = floor(m1 * SimpleUniform());
    s20 = floor(m2 * SimpleUniform());
    s21 = floor(m2 * SimpleUniform());
    s22 = floor(m2 * SimpleUniform());
    s23 = floor(m2 * SimpleUniform());
    s24 = floor(m2 * SimpleUniform());
  }
  /* Component 1 */
  p1 = a12 * s13 - a15n * s10;
  if (p1 > 0.0)
    p1 -= a14 * m1;
  p1 += a14 * s11;
  k = long(p1 / m1);
  p1 -= k * m1;
  if (p1 < 0.0)
    p1 += m1;
  s10 = s11;
  s11 = s12;
  s12 = s13;
  s13 = s14;
  s14 = p1;
  /* Component 2 */
  p2 = a21 * s24 - a25n * s20;
  if (p2 > 0.0)
    p2 -= a23 * m2;
  p2 += a23 * s22;
  k  = long(p2 / m2);
  p2 -= k * m2;
  if (p2 < 0.0)
    p2 += m2;
  s20 = s21;
  s21 = s22;
  s22 = s23;
  s23 = s24;
  s24 = p2;
  /* Combination */
  if (p1 <= p2)
    d = ((p1 - p2 + m1) * norm);
  else
    d = ((p1 - p2) * norm);
  return d;
}

// Using the polar form of Box-Muller transformation.
vector<double> Gaussian(int n, double m, double s) {
  srandom(time(NULL));
  vector<double> nums;
  for (int i = 0; i < n / 2; ++i) {
    double x1, x2, w, y1, y2;
    do {
      x1 = 2.0 * ranf() - 1.0;
      x2 = 2.0 * ranf() - 1.0;
      w = x1 * x1 + x2 * x2;
    } while (w >= 1.0);
    w = sqrt((-2.0 * log(w)) / w);
    y1 = x1 * w;
    y2 = x2 * w;
    nums.push_back(y1);
    nums.push_back(y2);
  }
  for (int i = 0; i < nums.size(); ++i) {
    nums[i] = nums[i] * s + m;
  }
  return nums;
}

void GenABDat() {
  // For AB data, all reads are of same size.
  int total_pairs = cov * org.size() / (2 * (low_len - 1));
  static vector<double> insert_lens = Gaussian(total_pairs + 1, ins_len, std_dev);
  for (int i = 0; i < total_pairs; ++i) {
    
    bool strand = rand() % 2;
    const string& ref = strand ? org : orgrc;
    //    int lpos = rand() % (ref.size() - ins_len - std_dev);
    //    int lpos = rand() % (ref.size() - ins_len - 2 * std_dev);
    int lpos = ecuyer2() * (ref.size() - ins_len - 2 * std_dev);
    //    int dev = rand() % std_dev;
    //    if (rand() % 2) dev *= -1;
    int rpos = lpos + insert_lens[i] - low_len;
    if (rpos + low_len + 1 > ref.size()) {
      rpos = ref.size() - low_len - 1;
    }
    if (!strand) {  // RC
      swap(lpos, rpos);
    }
    //    lpos = rpos = 0;
    const string& lmar = ref.substr(lpos, low_len + 1);
    const string& rmar = ref.substr(rpos, low_len + 1);
    //    cout << UtilsString::Fasta2Solid(lmar) << endl;
    (*fpair) << ">a" << lpos << (strand ? '+' : '-') << endl;
    (*fpair) << AddErr2AB(lmar) << endl;
    //(*fpair) << lmar << endl;

    (*fpair) << ">b" << rpos << (strand ? '+' : '-') << endl;
    (*fpair) << AddErr2AB(rmar) << endl;
    //   (*fpair) << rmar << endl;
  }
}

void GenSeed(int seed_len, int num_seed) {
  for (int i = 0; i < num_seed; ++i) {
    bool strand = rand() % 2;
    const string& ref = strand ? org : orgrc;
    int pos = rand() % (org.size() - seed_len);
    (*fseed) << ">" << pos << (strand ? '+' : '-') << endl;
    (*fseed) << ref.substr(pos, seed_len) << endl;
    fstream* f = UtilsFile::OpenFile("out/seed." + UtilsString::Int2Str(i), "w");
    (*f) << ">" << pos << (strand ? '+' : '-') << endl;
    (*f) << ref.substr(pos, seed_len) << endl;
  }
}
