// This is used fix insertion/deletion and substitution errors in reads.

#include <iostream>
#include <fstream>
#include <stack>

#include "header.h"
#include "utils-string.h"

//using namespace std;

// These are used to fix insertion/deletion errors.
//--------------------------
const int LLEN = 6;
const int RLEN = 6;
// [left LLEN-mar][right RLEN-mar][ATGCN]-->N means 'none' here.
//int vote4indel[1 << (LLEN << 1)][1 << (RLEN << 1)][5];
//--------------------------

// This will be used to fix substition errors based on the frequency of k+1-mars.
//--------------------------
const int KLEN = 10;
int **lvote4subs;//[1 << (KLEN << 1)][4];
int **rvote4subs;//[1 << (KLEN << 1)][4];
//int** subsand[4]; //[4][1 << (KLEN << 1)][1 << (KLEN << 1)]
//--------------------------

void init() {
  lvote4subs = new int* [1 << (KLEN << 1)];
  rvote4subs = new int* [1 << (KLEN << 1)];
  for (int i = 0; i < (1 << (KLEN << 1)); ++i) {
    lvote4subs[i] = new int [4];
    rvote4subs[i] = new int [4];
  }
}

void finalize() {
  for (int i = 0; i < (1 << (KLEN << 1)); ++i) {
    delete [] lvote4subs[i];
    delete [] rvote4subs[i];
  }
  delete [] lvote4subs;
  delete [] rvote4subs;
}

bool cmp(const StrIntPair& a, const StrIntPair& b) {
  return a.second > b.second;
}

string getMostFreqChar(int l, int r) {
//   StrIntPairVec ivec;
//   ivec.push_back(StrIntPair("A", vote4indel[r][0]));
//   ivec.push_back(StrIntPair("T", vote4indel[r][1]));
//   ivec.push_back(StrIntPair("G", vote4indel[r][2]));
//   ivec.push_back(StrIntPair("C", vote4indel[r][3]));
//   ivec.push_back(StrIntPair("", vote4indel[r][4]));
//   sort(ivec.begin(), ivec.end(), cmp);
//   return ivec[0].first;
}

int main(int argc, char* argv[]) {
  if (argc != 2) {
    cout << "usage: shorty-read-cleaner read-file-name" << endl;
    exit(1);
  }
  init();
  fstream fin;
  fin.open(argv[1], ios::in);
  string str;
  while (fin >> str) {
    if (str[0] == '>') continue;
#ifdef COLOR_SPACE
    str = UtilsString::Fasta2Solid(str);
#endif  // COLOR_SPACE
    // This is to fix base x like [KLEN]x.
    for (int i = 0; i + KLEN < str.size(); ++i) {

      // Not doing it in an efficient way.
      int v = UtilsString::String2Int(str, i, KLEN);
      assert(str[i + KLEN] == 'A' || str[i + KLEN] == 'T' ||
	     str[i + KLEN] == 'G' || str[i + KLEN] == 'C');
      switch (str[i + KLEN]) {
      case 'A':
	++rvote4subs[v][0];
	break;
      case 'T':
	++rvote4subs[v][1];
	break;
      case 'G':
	++rvote4subs[v][2];
	break;
      case 'C':
	++rvote4subs[v][3];
	break;
      default:
	cout << "what's going on?" << endl;
	exit(1);
      }
    }

    // This is to fix base x like x[KLEN].
    for (int i = 1; i + KLEN <= str.size(); ++i) {
      int v = UtilsString::String2Int(str, i, KLEN);
      assert(str[i - 1] == 'A' || str[i - 1] == 'T' || str[i - 1] == 'G'
	     || str[i - 1] == 'C');
      switch (str[i - 1]) {
      case 'A':
	++lvote4subs[v][0];
	break;
      case 'T':
	++lvote4subs[v][1];
	break;
      case 'G':
	++lvote4subs[v][2];
	break;
      case 'C':
	++lvote4subs[v][3];
	break;
      default:
	cout << "what's going on?" << endl;
	exit(1);
      }
    }
  }
  fin.close();
  fstream tin(argv[1], ios::in);
  while (tin >> str) {
    if (str[0] == '>') {
      cout << str << endl;
      continue;
    }
    
#ifdef COLOR_SPACE
    char firstc = str[0];
    str = UtilsString::Fasta2Solid(str);
#endif  // COLOR_SPACE
    // This is a parallel string that will reflect the changes.
    string tstr = str;
    
    // Take the base which has maximum no. of mars contributing.
    IntVec votes;
    for (int i = 0; i < str.size(); ++i) {
      votes.push_back(1);
    }
    for (int i = 0; i + KLEN < str.size(); ++i) {
      int v = UtilsString::String2Int(str, i, KLEN);
      for (int j = 0; j < 4; ++j) {
	if (rvote4subs[v][j] > votes[i + KLEN]) {
	  votes[i + KLEN] = rvote4subs[v][j];
	  tstr[i + KLEN] = UtilsString::Int2Base(j);
	}
      }
    }
    for (int i = 1; i + KLEN <= str.size(); ++i) {
      int v = UtilsString::String2Int(str, i, KLEN);
      for (int j = 0; j < 4; ++j) {
	if (lvote4subs[v][j] > votes[i - 1]) {
	  votes[i - 1] = lvote4subs[v][j];
	  tstr[i - 1] = UtilsString::Int2Base(j);
	}
      }
    }
    str = tstr;
#ifdef COLOR_SPACE
    tstr = "";
    tstr.push_back(firstc);
    for (int i = 0; i < str.size(); ++i) {
      tstr.push_back(UtilsString::Base2Int(str[i]) + '0');
    }
    str = "";
    str.push_back(firstc);
    str += UtilsString::Solid2Fasta(tstr);
#endif  // COLOR_SPACE
    cout << str << endl;
  }
  tin.close();
  // CAUTION: we are not taking care of indel errors at this point.
  //     for (int i = 0; i + LLEN + RLEN <= str.size(); ++i) {
  //       int l = UtilsString::String2Int(str, i, LLEN);
  //       int r = UtilsString::String2Int(str, i + LLEN, RLEN);
  //       ++vote4indell[r][4];
  
  //       if (i + LLEN + RLEN + 1 > str.size()) continue;
  
  //       assert(str[i + LLEN] == 'A' || str[i + LLEN] == 'T' ||
  // 	     str[i + LLEN] == 'G' || str[i + LLEN] == 'C');
  //       r = UtilsString::String2Int(str, i + LLEN + 1, RLEN);
  //       switch(str[i + LLEN]) {
  //       case 'A':
  // 	++vote4indell[r][0];
  // 	break;
  //       case 'T':
  // 	++vote4indell[r][1];
  // 	break;
  //       case 'G':
  // 	++vote4indell[r][2];
  // 	break;
  //       case 'C':
  // 	++vote4indell[r][3];
  // 	break;
  //       default:
  // 	break;
  //       }
  //     }
  //   }
  //   fin.close();
  //   fin.open(argv[1], ios::in);
  //   while (fin >> str) {
  //     if (str[0] == '>') {
  //       cout << str << endl;
  //       continue;
  //     }
  
  //     // Handling deletion errors.
  //     stack<StrIntPair> ins;
  //     for (int i = 0; i + LLEN + RLEN <= str.size(); ++i) {
  //       int l = UtilsString::String2Int(str, i, LLEN);
  //       int r = UtilsString::String2Int(str, i + LLEN, RLEN);
  //       const string& t = getMostFreqChar(l, r);
  //       ins.push(StrIntPair(t, i + LLEN));
  //       //str.insert(i + LLEN, t);
  //       //      if (t.size()) break; // Allow only one at this time.
  //     }
  //     int prev = -1;
  //     while (ins.size()) {
  //       StrIntPair si = ins.top();
  //       ins.pop();
  //       if (si.second == prev || si.second == prev - 1) continue;
  //       str.insert(si.second, si.first);
  //       prev = si.second;
  //     }
  // //     // Handling insertion/substitution errors.
  // //     for (int i = 0; i + LLEN + RLEN + 1 <= str.size(); ++i) {
  // //       int l = UtilsString::String2Int(str, i, LLEN);
  // //       int r = UtilsString::String2Int(str, i + LLEN + 1, RLEN);
  // //       const string& a = getMostFreqChar(l, r);
  // //       if (a.size()) str[i + LLEN] = a[0];
  // //       else str.erase(i + LLEN, 1);
  // //     }
  //    cout << str << endl;
  finalize();
  return 0;
}
