// This is a standalone program to take geography information of contigs and
// use them to merge/order them.

#include <iostream>
#include <fstream>

#include "contig.h"
#include "global.h"
#include "macros.h"
#include "overlap-graph.h"
#include "read-processor.h"
#include "shorty-string.h"
#include "utils-file.h"
#include "utils-math.h"
#include "utils-string.h"

const string PREFIX = "out/";

ReadProcessor* rp;

// Used by the graph
//====================
bool* visited;
bool* obsolete;
IntTripleVec* adj;
int* lclen;
IntPairVec* lpath;
int total_nodes;
int* node_size;
//====================

struct SortReadMap {
  bool operator () (const IntIntMtTriple& a, const IntIntMtTriple& b) {
    if (a.second == b.second) {
      if (a.first == b.first) {
	return a.third < b.third;
      } else {
	return a.first < b.first;
      }
    }
    return a.second < b.second;
  }
};

inline bool eq_read_map(const IntIntMtTriple& a, const IntIntMtTriple& b) {
  return a.first == b.first && a.second == b.second && a.third == b.third;
}

inline bool eq_adj_id(const IntTriple& a, const IntTriple& b) { 
  return a.first == b.first;
}

inline bool lt_adj_id(const IntTriple& a, const IntTriple& b) {
  return a.first < b.first;
}

inline bool lt_adj_bundle(const IntTriple& a, const IntTriple& b) {
  return a.third < b.third;
}

inline bool gt_adj(const IntPair& a, const IntPair& b) {
  return a.first > b.first;
}

inline bool gt_contig_lens(const IntTriple& a, const IntTriple& b) {
  if (a.first == b.first) return a.third > b.third;
  return a.first > b.first;
}

inline bool lt_pair(const IntPair& a, const IntPair& b) {
  if (a.first == b.first) {
    return a.second < b.second;
  }
  return a.first < b.first;
}

inline bool lt_pair1(const pair<int, bool>& a, const pair<int, bool>& b) {
  return a.first < b.first;
}

inline bool eq_pair(const IntPair& a, const IntPair& b) {
  return a.first == b.first && a.second == b.second;
}

void PrintReadMapping(const IntIntMtTripleVec& x) {
  for (int i = 0; i < x.size(); ++i) {
    //cout << x[i].first << " ";
    cout << "<" << x[i].first << "," << x[i].second << "," << (int)x[i].third << ">";
  }
  cout << endl << "============\n";
}

void ChimeraTest(const IntIntMtTripleVec& reads, int id, IntVec* cut_points) {
  cout << "##" << id << endl; 
  cut_points->clear();
  if (lclen[id] < Global::INSERT_LEN - 3 * Global::INSERT_LEN_DEV) {
    cut_points->push_back(0);
    cut_points->push_back(lclen[id]);
    return;
  }
  //   IntIntMtTripleVec reads;
  //   for (int i = 0; i < lpath[id].size(); ++i) {
  //     int x = lpath[id][i].first;
  //     int y = lpath[id][i].second;
  //     for (int j = 0; j < contigs[x].read_map.size(); ++j) {
  //       reads.push_back(IntIntMtTriple(contigs[x].read_map[j].first,
  // 				     contigs[x].read_map[j].second + y,
  // 				     contigs[x].read_map[j].third));
  //     }
  //   }
  //   sort(reads.begin(), reads.end(), SortReadMap());
  //   IntIntMtTripleVec::iterator new_end = unique(reads.begin(), reads.end(),
  // 					       eq_read_map);
  //   reads.erase(new_end, reads.end());
  IntPairVec l, r;
  for (int i = 0; i < reads.size(); ++i) {
    const IntIntMtTriple& x = reads[i];
    // assert(x.second >= 0);
    if (x.third == LMAR || x.third == LMARRC) {
      l.push_back(IntPair(x.first, x.second));
    } else {
      r.push_back(IntPair(x.first, x.second));
    }
  }
  sort(l.begin(), l.end(), lt_pair);
  sort(r.begin(), r.end(), lt_pair);
  IntPairVec::iterator new_end_pos = unique(l.begin(), l.end(), eq_pair);
  l.erase(new_end_pos, l.end());
  new_end_pos = unique(r.begin(), r.end(), eq_pair);
  r.erase(new_end_pos, r.end());
  vector<pair<int, bool> > event_points;
  int pl = 0;
  int pr = 0;
  while (pl < l.size() && pr < r.size()) {
    if (l[pl].first == r[pr].first) {
      int left = min(l[pl].second, r[pr].second);
      int right = max(l[pl].second, r[pr].second);

      // CAUTION
      // Not a correct estimate if the read lengths are different. Easy to fix.
      if (right - left <= Global::INSERT_LEN + 3 * Global::INSERT_LEN_DEV &&
	  right - left >= Global::INSERT_LEN - 3 * Global::INSERT_LEN_DEV) {
	event_points.push_back(pair<int, bool>(left, true));
	event_points.push_back(pair<int, bool>(right, false));
      }
      ++pl;
      ++pr;
    } else if (l[pl].first < r[pr].first) {
      ++pl;
    } else {
      ++pr;
    }
  }
  sort(event_points.begin(), event_points.end(), lt_pair1);
  int depth = 0;
  IntVec depths;
  for (int i = 0; i < event_points.size(); ++i) {
    if (event_points[i].second) ++depth;
    else --depth;
    //    cout << "[" << event_points[i].first << "," << depth << "]";
    cout << event_points[i].first << " " << depth << endl;      
    //    if (depth <= 0) {
    //  return true;
    // }
    depths.push_back(depth);
  }
  if (depths.size() == 0) {
    cut_points->push_back(0);
    cut_points->push_back(lclen[id]);    
    return;
  }
  long double avg = 0;
  for (int i = 0; i < depths.size(); ++i) {
    avg += (long double)depths[i] / depths.size();
  }
  int cutline = (int)avg / 2;
  cout << "cut line: " << cutline << endl;
  bool up = true;
  for (int i = 0; i < depths.size(); ++i) {
    if (up) { // looking for the up point.
      if (depths[i] > cutline) {
	up = false;
	cut_points->push_back(event_points[i].first);
      }
    } else {  // looking for the down point.
      if (depths[i] <= cutline) {
	up = true;
	cut_points->push_back(event_points[i].first);
      }
    }
  }
}

void BuildGraph() {
  visited = new bool[total_nodes];
  obsolete = new bool[total_nodes];
  lpath = new IntPairVec[total_nodes];
  lclen = new int[total_nodes];
  for (int i = 0; i < total_nodes; ++i) {
    visited[i] = false;
    obsolete[i] = false;
    lclen[i] = node_size[i];
    lpath[i].push_back(IntPair(i, 0));
  }
  for (int i = 0; i < total_nodes; ++i) {
    sort(adj[i].begin(), adj[i].end(), lt_adj_id);
    IntTripleVec::iterator new_end_pos = unique(adj[i].begin(),
						adj[i].end(), eq_adj_id);
    adj[i].erase(new_end_pos, adj[i].end());
    sort(adj[i].begin(), adj[i].end(), lt_adj_bundle);
  }
  cout << "graph building finished\n";
  //PrintAdj();
}

int total_unvisited;

void Traverse(int curr) {
  stack<IntPair> nodes;  // root + depth
  stack<IntPair> tnodes;  // elegth+ poffset
  nodes.push(IntPair(curr, 0));
  tnodes.push(IntPair(node_size[curr], 0));
  IntPairVec path;
  while (nodes.size()) {
    int root = nodes.top().first;
    int depth = nodes.top().second;
    int elength = tnodes.top().first;
    int poffset = tnodes.top().second;
    nodes.pop();
    tnodes.pop();
    if (visited[root]) {  // Due to cycles, it was already visited.
      continue;
    }
    visited[root] = true;
    --total_unvisited;
    //    cout << "unvisited: " << total_unvisited << endl;
    //cout << "visited " << root << endl;
    if (path.size() >= depth + 1) {
      path[depth] = IntPair(root, poffset);
    } else {
      path.push_back(IntPair(root, poffset));
    }
    bool any  = false;
    for (int i = 0; i < adj[root].size(); ++i) {
      int j = adj[root][i].first;
      if (visited[j]) {
	int add = (node_size[root] - adj[root][i].second) > lclen[j] ? 0
	  : (lclen[j] - node_size[root] + adj[root][i].second);
	int newlen = elength - poffset < node_size[root] + add ?
	  (poffset + node_size[root] + add) : elength;
	int start = path[0].first;
	if (newlen >= lclen[start]) {
	  //contigs[j].obsolete = false;
	  lclen[start] = newlen;
	  IntPairVec tpath; // Need this to avoid a cycle.
	  //lpath_[start].clear();
	  for (int k = 0; k <= depth; ++k) {
	    tpath.push_back(path[k]);
	  }
	  for (int k = 0; k < lpath[j].size(); ++k) {
	    tpath.push_back(IntPair(lpath[j][k].first,
				    adj[root][i].second + lpath[j][k].second +
				    poffset));
	  }
	  lpath[start] = tpath;
	}
	continue;
      }
      any = true;
      int add = (node_size[root] - adj[root][i].second) > node_size[j]
	? 0 : (node_size[j] - node_size[root] + adj[root][i].second);
      int newlen = elength - poffset < node_size[root] + add ? 
	(poffset + node_size[root] + add) : elength;
      nodes.push(IntPair(adj[root][i].first, depth + 1));
      tnodes.push(IntPair(newlen, poffset + adj[root][i].second));
      //break;  // We only want one path.
    }
    if (!any) {
      int start = path[0].first;
      if (elength > lclen[start] ||
	  (elength == lclen[start] && depth >= lpath[start].size())) {
	lclen[start] = elength;
	lpath[start].clear();
	for (int i = 0; i <= depth; ++i) {
	  lpath[start].push_back(path[i]);
	  //	  if (i) contigs[path[i].first].obsolete = true;
	}
      } else {
	// Find the node where this path diverted. That might well be 
	// considered a sucpicious node making a chimera.
	for (int i = 1; i <= depth && i < lpath[start].size(); ++i) {
	  int id = path[i].first;
	  if (id != lpath[start][i].first) {
	    //int clen = elength - path[i].second;
	    int soffset = path[i].second;
	    int clen = node_size[id];
	    for (int j = i + 1; j <= depth; ++j) {
	      clen = max(clen, path[j].second - soffset +
			 node_size[path[j].first]);
	    }
	    if (clen > lclen[id]) {
	      lclen[id] = clen;
	      lpath[id].clear();
	      for (; i <= depth; ++i) {
		lpath[id].push_back(IntPair(path[i].first,
					    path[i].second - soffset));
	      }
	    }
	    break;
	  }
	}
      }
    }
  }
}

void VisitGraph() {
  total_unvisited = total_nodes;
  BuildGraph();
  for (int i = 0; i < total_nodes; ++i) {
    if (visited[i]) continue;
    Traverse(i);
    //    cout << "traversed " << i << endl;
  }
}

void PrintConfigInfo() {
  cout << "running with the following configurations:\n"
       << "overlap between contigs\t\t\t" << Global::MIN_OVERLAP << endl
       << "min score for overlap\t\t\t" << Global::MIN_SCORE << endl
       << "min contig len\t\t\t\t" << Global::DESIRED_CONTIG_LEN << endl;
}

// void AnalyzeContig(int id) {
//   cout << lpath[id][0].first;
//   for (int i = 1; i < lpath[id].size(); ++i) {
//     int x = lpath[id][i - 1].first;
//     int y = lpath[id][i].first;
//     int offseta, bundle;
//     bool overlap = OverlapTest(x, y, &offseta, &bundle);
//     if (!overlap) {
//       //      cout << endl << x << "!o" << y << endl;
//       //exit(1);
//       cout << "(-)";
//     } else {
//       cout << "(" << bundle << ")";
//     }
//     cout << y;
//   }
//   cout << endl;
// }

int main(int argc, char* argv[]) {
  // Read the arguments.
  int c;
  while ((c = getopt(argc, argv, "o:l:")) != -1) {
    switch (c) {
    case 'o':
      Global::MIN_OVERLAP = atoi(optarg);
      Global::MIN_SCORE = 4 * Global::MIN_OVERLAP -
	min((int)(0.25 * 4 * Global::MIN_OVERLAP), 25);
      break;
    case 'l':
      Global::DESIRED_CONTIG_LEN = atoi(optarg);
      break;
    default:
      cout << "unknown option!!!\n";
      break;
    }
  }
  PrintConfigInfo();
  const string& read_file = argv[optind];
  rp = new ReadProcessor(read_file);
  total_nodes = rp->TotalPairs() * MAR_TYPES;
  node_size = new int[total_nodes];
  for (int i = 0; i < total_nodes; ++i) {
    node_size[i] = rp->GetMarLen(i % rp->TotalPairs(),
				 (mar_types)(i / rp->TotalPairs()));
  }
  const string& ext = argv[optind + 1];
  cout << "using output extension: " << ext << endl;
  int input_file = atoi(argv[optind + 2]);
  cout << "total seed: " << input_file << endl;
  adj = new IntTripleVec[total_nodes];
  for (int i = 0; i < input_file; ++i) {
    const string& geo_info = argv[optind + 2 * i + 3];
    const string& read_info = argv[optind + 2 * i + 4];
    cout << "using geom file: " << geo_info << endl;
    cout << "using geo file: " << read_info << endl;
    fstream* geo = UtilsFile::OpenFile(geo_info, "r");
    int sid, id;
    ShortyString contig;
    *geo >> sid >> id;
    *geo >> contig >> id;
    *geo >> id;
    *geo >> id;
    contig.ReadVotesFromFile(*geo);
    int slen = contig.size();
    fstream* read_mapping = UtilsFile::OpenFile(read_info, "r");
    while (*read_mapping >> id) {
      cout << "processing: " << id << endl;
      int total_reads;
      *read_mapping >> total_reads;
      IntIntMtTripleVec read_map;
      while (total_reads--) {
	int rid, roffset, mt;
	*read_mapping >> rid >> roffset >> mt;
	read_map.push_back(IntIntMtTriple(rid, roffset, (mar_types)mt));
      }
      sort(read_map.begin(), read_map.end(), SortReadMap());
      
      // Construct adjacency information from here.
      for (int k = 0; k < read_map.size() - 1; ++k) {
	int x = rp->TotalPairs() * read_map[k].third + read_map[k].first;
	int j = k + 1;
	//	int to = read_map[j].second;
	//while (j < read_map.size() && read_map[j].second == to) {
	int y = rp->TotalPairs() * read_map[j].third + read_map[j].first;
	int o = read_map[j].second - read_map[k].second;
	int b = node_size[x] - o;
	//if (b <= 0) { //Global::MIN_OVERLAP) {
	  //	  cout << read_map[j].first << " " << read_map[j].second << endl;
	  //cout << read_map[k].first << " " << read_map[k].second << endl;
	  //cout << "x: " << x << " " << "o: " << o << " " << "b: " << b << endl;
	  //exit(0);
	  //continue;
	//	}
	assert(b > 0);
	adj[x].push_back(IntTriple(y, o, b));
	  //++j;
	  //	}
      }
      break;
    }
    UtilsFile::CloseFile(geo);
    UtilsFile::CloseFile(read_mapping);
    break;
  }
  VisitGraph();
  
  IntTripleVec contig_lens;
  for (int i = 0; i < total_nodes; ++i) {
    //if (contigs[i].obsolete) continue;
    
    // CAUTION: somewhere we have a miscalculation. We are always getting one
    // more than the actual lclen.
    //--lclen[i];
    contig_lens.push_back(IntTriple(lclen[i], i, lpath[i].size()));
  }
  cout << "size: " << contig_lens.size() << endl;
  sort(contig_lens.begin(), contig_lens.end(), gt_contig_lens);
  //  cout << "total: " << contig_lens.size() << endl;
  for (int t = 0; t < contig_lens.size(); ++t) {
    int i = contig_lens[t].second;
    // if (contigs[i].obsolete) continue;
    if (obsolete[i]) continue;
    if (lclen[i] < Global::DESIRED_CONTIG_LEN) break;
    cout << ">len: " << lclen[i] << "(" << lpath[i].size() << ")" << endl;
    int l = 0;
    //int total_len = 0;
    for (int k = 0; k < lpath[i].size(); ++k) {
      if (obsolete[lpath[i][k].first]) continue;
      //      total_len += contigs[lpath[i][k].first].size();
      //      cout << "<" << lpath[i][k].first << "." << lpath[i][k].second << "."
      //	   << contigs[lpath[i][k].first].size() << ">";
      lpath[i][l++] = lpath[i][k];
      if (lpath[i][k].first != i) obsolete[lpath[i][k].first] = true;
    }
    //    cout << endl;
    //cout << ">total len: " << total_len << endl;
    lpath[i].resize(l);
  }
  fstream* econtigs = UtilsFile::OpenFile(PREFIX + "contigs-enriched" + ext,
 					  "w");
  fstream* egeom = UtilsFile::OpenFile(PREFIX + "geom-enriched" + ext, "w");
  fstream* eread_info = UtilsFile::OpenFile(PREFIX + "geo-enriched" + ext, "w");
  int total_econtig = 0;
  cout << "starting read merge\n";
  for (int k = 0; k < contig_lens.size(); ++k) {
    int i = contig_lens[k].second;
    if (obsolete[i]) continue;
    IntIntMtTripleVec read_map;
    if (lclen[i] < Global::DESIRED_CONTIG_LEN) break;
    for (int j = 0; j < lpath[i].size(); ++j) {
      read_map.push_back(IntIntMtTriple(lpath[i][j].first % rp->TotalPairs(),
					lpath[i][j].second,
					(mar_types)(lpath[i][j].first /
						    rp->TotalPairs())));
    }
    ShortyString str;
    UtilsString::MergeReads(read_map, lclen[i], rp, &str);
    sort(read_map.begin(), read_map.end(), SortReadMap());
    IntIntMtTripleVec::iterator new_end_pos = unique(read_map.begin(),
 						     read_map.end(),
 						     eq_read_map);
    read_map.erase(new_end_pos, read_map.end());
    IntVec cut_points;
    ChimeraTest(read_map, i, &cut_points);
    assert (cut_points.size() % 2 == 0);
    cut_points[0] = 0;
    cut_points[cut_points.size() - 1] = lclen[i];
    int prindex = 0;
    for (int j = 1; j < cut_points.size(); ++j) {
      if (j % 2 == 0) {
	cout << "throwing a segment of size: "
	     << cut_points[j] - cut_points[j - 1] << endl;
	continue; // This is a bad portion.
      }
      while (prindex < read_map.size() &&
             read_map[prindex].second
             <= cut_points[j - 1]) {
        ++prindex;
      }
      int rstart = prindex;
      while (prindex < read_map.size() && read_map[prindex].second +
	     node_size[prindex] <= cut_points[j]) {
 	++prindex;
      }
      int rend = prindex - 1;
      ShortyString pstr = str.substr(cut_points[j - 1],
 				     cut_points[j] - cut_points[j - 1]);
      if (pstr.size() < Global::DESIRED_CONTIG_LEN) continue;
      // Replace the trailing dots.
      int l;
      for (l = pstr.size() - 1; l >= 0; --l) {
 	if (pstr[l] != '.') break;
      }
      pstr = pstr.substr(0, l + 1);
      if (pstr.size() == 0) continue;
      for (l = 0; l < pstr.size(); ++l) {
  	if (pstr[l] == '.') break;
      }
      if (l < pstr.size()) continue;
      *econtigs << ">contig_" << i << "_" << j - 1 << endl << pstr << endl; 
      *egeom << ">contig_" << total_econtig << endl << total_econtig << " "
	     << pstr << " -1 0 -1" << endl;
      pstr.PrintScores(*egeom);
      *eread_info << total_econtig << " " << rend - rstart + 1 << endl;
      while (rstart <= rend) {
 	*eread_info << read_map[rstart].first << " "
 		    << read_map[rstart].second - cut_points[j - 1]
 		    << " " << read_map[rstart].third << " ";
 	++rstart;
      }
      *eread_info << endl;
      ++total_econtig;
    }
  }
  UtilsFile::CloseFile(econtigs);
  UtilsFile::CloseFile(egeom);
  UtilsFile::CloseFile(eread_info);
  return 0;
}
