#include <fstream>
#include <iostream>

#include "compact-trie.h"
//#include "constants.h"
#include "global.h"
#include "raw-read.h"
#include "read-processor.h"
#include "trie.h"
#include "utils-file.h"
#include "utils-string.h"

using namespace std;

int ReadProcessor::total_;
int ReadProcessor::cov_;
int* ReadProcessor::used_;
char* ReadProcessor::num_used_;
//int ReadProcessor::prev_used_[NUM_PAIR];
//vector<const char*> ReadProcessor::all_reads_c;
vector<string> ReadProcessor::all_reads_;
vector<RawRead> ReadProcessor::raw_read_;

ReadProcessor::ReadProcessor(const string& path) {
  //fill(prev_used_, prev_used_ + NUM_PAIR, -1);
  //  trie_ = new Trie;
  trie_ = new TrieNode;
  // CAUTION: check it out....
  readsVect = this;

  // This file contains all the raw-paired reads.
  fstream* pairs = UtilsFile::OpenFile(path, "r");
  
  // We are assumming all reads are preceeded by an id or some other info.
  total_ = 0;
  cov_ = 0;
  string ida, idb;
  while ((*pairs) >> ida) {
    string lmar, rmar;
    (*pairs) >> lmar >> idb >> rmar;
    //    if (rand() % 2 != 0) continue;
    cov_ += lmar.size();
    cov_ += rmar.size();
    
    // Ignore the not-useful pairs.
    //if (lmar.size() < MIN_READ_LEN || rmar.size() < MIN_READ_LEN) continue;

#ifdef COMPRESS
    lmar = UtilsString::CompressStr(lmar);
    rmar = UtilsString::CompressStr(rmar);
#endif  // COMPRESS
#ifndef COLOR_SPACE
    const string& lmarrc = UtilsString::ReverseComplement(lmar);
    const string& rmarrc = UtilsString::ReverseComplement(rmar);
#endif  // COLOR_SPACE
#ifdef COLOR_SPACE
    //    lmar = UtilsString::Fasta2Solid(lmar);
    // rmar = UtilsString::Fasta2Solid(rmar);
    lmar = lmar.substr(1, lmar.size() - 1);
    rmar = rmar.substr(1, rmar.size() - 1);
    const string& lmarrc = UtilsString::Reverse(lmar);
    const string& rmarrc = UtilsString::Reverse(rmar);
#endif  // COLOR_SPACE
    int l = LocateRead(lmar);
    int lrc = LocateRead(lmarrc);
    int r = LocateRead(rmar);
    int rrc = LocateRead(rmarrc);
    //    raw_read_.push_back(RawRead(lmar, rmar, ida, idb, lqual, rqual, total_++));
    raw_read_.push_back(RawRead(total_, l, lrc, r, rrc));
    //    trie_->Insert(GetMar(total_, LMAR), total_, LMAR);
    //trie_->Insert(GetMar(total_, LMARRC), total_, LMARRC);
    //trie_->Insert(GetMar(total_, RMAR), total_, RMAR);
    //trie_->Insert(GetMar(total_, RMARRC), total_, RMARRC);
    trie_->Insert(total_, LMAR);
    trie_->Insert(total_, LMARRC);
    trie_->Insert(total_, RMAR);
    trie_->Insert(total_, RMARRC);
    total_++;
  }
  used_ = new int[total_];
  num_used_ = new char[total_];
  fill(used_, used_ + total_, -1);
  fill(num_used_, num_used_ + total_, 0);
  Global::TOTAL_PAIRS = total_;
#ifdef DEBUG
  cout << "total # of pairs(unique reads): " << total_ << "("
       << all_reads_.size() << ")" << endl;
#endif  // DEBUG
  UtilsFile::CloseFile(pairs);
  //  cov_ /= Global::REF_SEQ_LEN;
  //cout << "coverage: " << cov_ << endl;
}

int ReadProcessor::LocateRead(const string& read) const {
  const IntMtPair& t = trie_->Search(read);
  int p = t.first;
  if (p == -1) {
    p = all_reads_.size();
    all_reads_.push_back(read);
    //const char* str = read.c_str();
    //    strcpy(str, read.c_str());
    //   all_reads_c.push_back(str);    
  } else {
    //    cout << read << " matches " << GetMar(p, t.second) << endl;
    p = raw_read_[p].Mar(t.second);
  }
  return p;
}

ReadProcessor::~ReadProcessor() {
  delete trie_;
  delete [] used_;
  delete [] num_used_;
}

IntIntMtTripleVec ReadProcessor::GetRecords(const string& str) const {
  
  // return trie_->Search(str, offset, SUB_2MAP_SEED);
  return trie_->searchWithErrorWholeSeed(str, Global::SUB_2MAP_SEED, GetMarLen(0, LMAR));
}

const string& ReadProcessor::GetMar(int id, mar_types mt) const {
  return all_reads_[raw_read_[id].Mar(mt)];
}

int ReadProcessor::GetMarLen(int id, mar_types mt) const {
  return all_reads_[raw_read_[id].Mar(mt)].size();
}

int ReadProcessor::TotalPairs() const { 
  return raw_read_.size();
}

void ReadProcessor::Debug() const {
  //  trie_->Insert("ATGCTTG", 1, LMAR);
  //  cout << "found at: " << trie_->Search("ATGC") << endl;
}
