#ifndef UTILS_MATH_H
#define UTILS_MATH_H

#include <cmath>
#include <numeric>

#include "header.h"

using namespace std;

// Class to host some math utility functions.
class UtilsMath {
 public:
  // It will find the element in the ith percentile in 'c'. Note that it will
  // change the contents of 'c'. Also right now it may only work for containers
  // that have data of type for which the built in comparators work.
  template<class RetType, class ContainerType>
    static RetType IthPercentile(int i, ContainerType* c) {
    assert(c->size());
    double p = i / 100.0;
    
    // 'nth_element' works in linear time.
    nth_element(c->begin(), c->begin() + (int)(c->size() * p), c->end());
    return *(c->begin(), c->begin() + (int)(c->size() * p));
  }

  template<class T>
    static T Abs(const T& a) 
    { return (a < 0) ? (-a) : a; }
  template<class ContainerType>
    static double Average(const ContainerType& v) 
    { return accumulate(v.begin(), v.end(), 0) / (double)v.size(); }

  static int MedianOfIntervals(const IntPairVec& v, int* mn, int* mx);

  template<class ContainerType>
    static double StandardDev(const ContainerType& v) {
    assert(v.size());
    double avg = Average<ContainerType>(v);
    long double sd = 0;
    for (int i = 0; i < v.size(); ++i) {
      sd += (avg - v[i]) * (avg - v[i]) / v.size();
    }
    return sqrt(sd);
  }

  // Returns 4^x.
  static int PowerOf4(int x) { return (1 << (x << 1)); }
 private:

  // This is used by 'MedianOfIntervals'.
  typedef struct median_event_point {
    bool left; // left or right point?
    int id; // id of interval.
    int x; // value of the point.
    
    bool operator < (const median_event_point& p) const {
      return x < p.x;
    }
  };

};

#endif  //  UTILS_MATH_H
